#!/usr/bin/env ratlperl
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2015, 2015.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp

##############################################################################
#  findUnwantedDependency - This script reviews a given stream. For each read 
#  only composite component it compares the latest baseline to the foundation,
#  looking for any dependencies that appear to have been removed (but might 
#  have left a "phantom" dependency in the stream). If dependencies appear 
#  to have been removed, it prints a sample mkbl ... -ddepends command that 
#  can be used to remove "phantoms".
##############################################################################

use Getopt::Long;

if ($^O =~ "Win32") {
    if ($ENV{'CLEARCASEHOME'}){
	$atria_home = $ENV{'CLEARCASEHOME'};
    } else {
	$atria_home = $ENV{'ATRIAHOME'};
    }
    die("Need CLEARCASEHOME or ATRIAHOME variables set to find tools\n") if (!$atria_home);
    $ps = "\\";
    $CT = "\"$atria_home${ps}bin${ps}cleartool\"";
    $UCMUTIL = "\"$atria_home${ps}etc${ps}utils${ps}ucmutil\"";
} else {
    $atria_home = "/opt/rational/clearcase";
    $ps = '/';
    $CT = "\"$atria_home${ps}bin${ps}cleartool\"";
    $UCMUTIL = "\"$atria_home${ps}etc${ps}utils${ps}ucmutil\"";
}

#----------------------------
# Some global hashes
#----------------------------
my %lastMakeBaseline    = ();  # hash for last baselines 

#----------------------------
# Some global arrays
#----------------------------
# my @blHierarchy = ();
my @streamBLs = ();

#----------------------------
# Some global scalars
#----------------------------
# my $debug = $ENV{ 'CLEARCASE_TRACE_TRIGGERS' };
my $debug	= "";
my $stream	= "";

# Get options and command line arguments
&getOpts();

# generate list of latest composite baselines created in this stream
@streamBLs = `$CT lsbl -stream $stream -fmt "\%[component]Xp \%Xn\\n"`;
print @streamBLs if $debug;
foreach my $BL (@streamBLs)
{
	my ($comp,$bl) = split(' ',$BL);
	$lastMakeBaseline{$comp} = $bl;			# Putting them in a hash will save only the last one
}
	
# create lookup hash for components actually in the stream configuration
foreach my $component (split(' ',`$CT lsstream -fmt "%[components]Xp" $stream`))
{
	$stream::config_components{substr($component,10)} = 1;
}

# find current stream foundation, per component, and compare
foreach my $comp (keys %lastMakeBaseline)
{
	# ignore some components ? "modifiable by this stream" 
    # some criteria you may want to apply even before running dump_timeline
	next unless (my $comp_root_dir = `$CT lscomp -fmt "%[root_dir]p" $comp`) eq '';	# Only looking at rootless components

	print ("$UCMUTIL dump_timeline -stream $stream -comp $comp\n") if $debug;
	my @timeline = `$UCMUTIL dump_timeline -stream $stream -comp $comp`;
	my @foundations = grep(/^rebase/,@timeline);
	my @baselines = grep(/^baseline/,@timeline);

	print ("foundations @foundations\n") if $debug;
	# ignore some components ? "modifiable by this stream" 
	# Example: Ignore if the stream is still on its original foundation for this component
	#    (Then any extra dependencies MUST have been deliberately added)
	next if @foundations < 2;
	
	my (undef,$currFoundation,$rebaseTimestamp) = split(' ',$foundations[-1]); 
	my (undef,$latestBaseline,$baselineTimestamp) = split(' ',$baselines[-1]); 
	print ("currFoundation $currFoundation  rebaseTimestamp $rebaseTimestamp\n") if $debug;
	print ("latestBaseline $latestBaseline  baselineTimestamp $baselineTimestamp\n") if $debug;
	print ("Warning: detected multiple timelines for $comp\n") if $latestBaseline ne substr($lastMakeBaseline{$comp},9);
	if (my @droppedDependencies = &findDroppedDependencies ($latestBaseline,$currFoundation))
	{
		# purge from the list any components that are no longer in the stream config
		my @ddepends = ();
		foreach my $dependency (@droppedDependencies)
		{
			push(@ddepends,$dependency) if exists $stream::config_components{$dependency};
		}
		next unless @ddepends;
		print "It looks like baseline dependencies in this stream, for $comp\n"
			. "include dependencies not present in its foundation.\n"
			. "Consider removing them with the following command:\n";
		printf ("  cleartool mkbl -c \"ddep %2\$s\" -component %s -ddepends %s [baseline name]\n",$comp,join(',',@ddepends));
		print "\n";
	}
}
exit 0;

sub findDroppedDependencies
###############################################################
#  Args:	1. Baseline considered "older"
#			1. Baseline considered "newer"
#  Output:	List of components that are present in the first but not the second
#  Errors:	Dies if arguments are not two baselines of the same component
###############################################################
{
	my ($currBL,$newBL) = @_;
	my $diffblcmd =  "$CT diffbl -baselines -nrecurse $currBL $newBL";
	my @diffBL = `$diffblcmd`;
	print ("$diffblcmd\n",@diffBL) if $debug;
	warn "Could not execute:  '$diffblcmd'" if $?; 
	my @droppedDependencies;
	for (my $i = 4; $i < @diffBL; $i++)
	{
		my ($arrows_i,$bl_i,$comp_i) = split(/[ \(\)]+/,$diffBL[$i]);
		next unless $arrows_i eq '<<';
		my ($arrows_j,$bl_j,$comp_j) = split(/[ \(\)]+/,$diffBL[$i+1]);
		($i++ and next) if ($arrows_j eq '>>' and $comp_i eq $comp_j);
		push (@droppedDependencies,$comp_i);
	}
	return @droppedDependencies;
}

sub getOpts {
        my $opt = {};
        Getopt::Long::Configure( "ignore_case", "auto_abbrev" );
        my $ret_val = Getopt::Long::GetOptions(
                $opt, "-debug!", "-help!"
        );

#        $long              = 1                  if ( $opt->{'long'} );
        $debug             = $opt->{'debug'}    if ( $opt->{'debug'} );

        my $usage = "
$0 compares latest baselines to foundation baselines 
for the read only composite components in a given stream. If it appears that dependencies have been 
removed it prints an example mkbl command that could be used to remove the phantom dependencies from 
the stream.

$0 [-help] [-debug] <stream>

    -help               print this usage
    -debug         	Debug mode for the script    

Example: $0 release_3.0_int@/vobs/projects
";

        die "$usage" if ( $ret_val != 1 );

        if ( $opt->{'help'} ) {
                print $usage;
                exit 0;
        }
#  Stream should be only remaining argument after option processing
	$stream 	= $ARGV[0];
	print "\$stream: $stream\n" if $debug;
	if ( $stream eq "" ) {
     		print "Must specify stream\n";
     		die "$usage";
    	}


}

